# type: ignore[union-attr]
from typing import Optional, Iterable, Union, cast, Iterator, Dict, Sequence, TYPE_CHECKING, Any, List, Mapping
from warnings import warn
import numpy as np
import pandas as pd
from monetdbe.connection import Connection, Description
from monetdbe.exceptions import ProgrammingError, InterfaceError
from monetdbe.formatting import format_query, strip_split_and_clean, parameters_type
from monetdbe.monetize import monet_identifier_escape, convert

if TYPE_CHECKING:
    from monetdbe.row import Row

paramstyles = {"qmark", "numeric", "named", "format", "pyformat"}


def _pandas_to_numpy_dict(df: pd.DataFrame) -> Dict[str, np.ndarray]:
    return {label: np.array(column) for label, column in df.iteritems()}  # type: ignore


class Cursor:
    lastrowid = 0

    def __init__(self, con: 'Connection'):

        if not isinstance(con, Connection):
            raise TypeError

        # changing array size has no effect for monetdbe
        self.arraysize = 1

        self.connection: Optional[Connection] = con
        self.rowcount = -1
        self.prepare_id: Optional[int] = None
        self.description: Optional[Description] = None
        self.row_factory = None

        self._fetch_generator: Optional[Iterator['Row']] = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def __del__(self):
        self.close()

    def _set_description(self):
        self.description = self.connection.get_description()

    def __iter__(self) -> Iterator[Union['Row', Sequence[Any]]]:
        # we import this late, otherwise the whole monetdbe project is unimportable
        # if we don't have access to monetdbe shared library
        from monetdbe._cffi.convert import extract
        from monetdbe._cffi.internal import result_fetch

        self._check_connection()

        if not self.connection.result:
            raise StopIteration

        columns = list(map(lambda x: result_fetch(self.connection.result, x),
                           range(self.connection.result.ncols)))  # type: ignore[union-attr]
        for r in range(self.connection.result.nrows):
            row = tuple(extract(rcol, r, self.connection.text_factory) for rcol in columns)
            if self.connection.row_factory:
                yield self.connection.row_factory(cur=self, row=row)
            elif self.row_factory:  # Sqlite backwards compatibly
                yield self.row_factory(self, row)
            else:
                yield row

    def _check_connection(self):
        """
        Check if we are attached to the lower level interface

        Raises:
            ProgrammingError: if no lower level interface is attached
        """
        if not hasattr(self, 'connection') or not self.connection:
            raise ProgrammingError("no connection to lower level database available")

    def _check_result(self) -> None:
        """
        Check if an operation has been executed and a result is available.

        Raises:
            ProgrammingError: if no result is available.
        """
        if not self.connection.result:
            raise ProgrammingError("fetching data but no query executed")

    def _execute_python(self, operation: str, parameters: parameters_type = None) -> 'Cursor':
        """
        Execute operation with Python based statement preparation.

        Args:
            operation: the query you want to execute
            parameters: an optional iterable containing arguments for the operation.

        Returns:
            the cursor object itself

        Raises:
            OperationalError: if the execution failed
        """
        self._check_connection()

        self.description = None  # which will be set later in fetchall

        self.connection.cleanup_result()

        splitted = strip_split_and_clean(operation)
        if len(splitted) == 0:
            raise ProgrammingError("Empty query")
        if len(splitted) > 1:
            raise ProgrammingError("Multiple queries in one execute() call")

        formatted = format_query(operation, parameters)
        self.connection.result, self.rowcount = self.connection.query(formatted, make_result=True)
        self.connection.total_changes += self.rowcount
        self._set_description()
        return self

    def _execute_monetdbe(self, operation: str, parameters: parameters_type = None):
        from monetdbe._cffi.internal import bind, execute
        self._check_connection()
        prepare_result = self.connection.prepare(operation)
        statement = prepare_result[0]

        self.connection.type_info = None
        if len(prepare_result) == 2:
            self.connection.type_info = prepare_result[1]

        if parameters:
            for index, parameter in enumerate(parameters):
                bind(statement, parameter, index, self.connection.type_info)
        self.connection.result, self.rowcount = execute(statement, make_result=True)
        self.connection.cleanup_statement(statement)
        self.connection.total_changes += self.rowcount
        self._set_description()
        return self

    def execute(
            self,
            operation: str,
            parameters: parameters_type = None,
            paramstyle: str = "qmark"
    ) -> 'Cursor':
        if paramstyle not in paramstyles:
            raise ValueError(f"Unknown paramstyle {paramstyle}")
        if (not parameters or isinstance(parameters, Sequence)) and paramstyle == "qmark":
            return self._execute_monetdbe(operation, parameters)
        return self._execute_python(operation, parameters)

    def executemany(self, operation: str, seq_of_parameters: Union[Iterator, Iterable[Iterable]]) -> 'Cursor':
        """
        Prepare a database operation (query or command) and then execute it against all parameter sequences or
        mappings found in the sequence seq_of_parameters.

        Args:
            operation: the SQL query to execute
            seq_of_parameters: An optional iterator or iterable containing an iterable of arguments
        """
        self._check_connection()
        self.description = None  # which will be set later in fetchall
        self.connection.cleanup_result()
        total_affected_rows = 0

        if operation[:6].lower().strip() == 'select':
            raise ProgrammingError("Don't use a SELECT statement with executemany()")

        if hasattr(seq_of_parameters, '__iter__'):
            iterator = iter(seq_of_parameters)
        else:
            iterator = seq_of_parameters  # type: ignore   # mypy gets confused here

        while True:
            try:
                parameters = next(iterator)
            except StopIteration:
                break

            formatted = format_query(operation, parameters)
            self.connection.result, affected_rows = self.connection.query(formatted, make_result=True)
            total_affected_rows += affected_rows

        self.rowcount = total_affected_rows
        self.connection.total_changes += total_affected_rows
        self._set_description()
        return self

    def close(self) -> None:
        """
        Shut down the connection.
        """
        if hasattr(self, 'connection') and self.connection and self.connection.result:
            self.connection.cleanup_result()
            self.connection.result = None
        self.connection = None

    def executescript(self, sql_script: str) -> None:
        """
        This is a nonstandard convenience and SQLite compatibility method for executing multiple SQL statements at once.
        It issues a COMMIT statement first, then executes the SQL script it gets as a parameter.

        Args:
            sql_script: A string containing one or more SQL statements, split by ;

        Raises:
            Error: If the previous call to .execute*() did not produce any result set or no call was issued yet.

        """
        self._check_connection()

        if not isinstance(sql_script, str):
            raise ValueError("script argument must be unicode.")

        for query in strip_split_and_clean(sql_script):
            self.execute(query)

    def create(self, table, values, schema=None):
        """
        Creates a table from a set of values or a pandas DataFrame.
        """
        # note: this is a backwards compatibility function with monetdblite
        self._check_connection()

        column_types = []

        if not isinstance(values, dict):
            values = _pandas_to_numpy_dict(values)
        else:
            vals = {}
            for tpl in values.items():
                if isinstance(tpl[1], np.ma.core.MaskedArray):
                    vals[tpl[0]] = tpl[1]
                else:
                    vals[tpl[0]] = np.array(tpl[1])
            values = vals
        if schema is None:
            schema = "sys"
        for key, value in values.items():
            arr = np.array(value)
            if arr.dtype == np.bool_:
                column_type = "BOOLEAN"
            elif arr.dtype == np.int8:
                column_type = 'TINYINT'
            elif arr.dtype == np.int16 or arr.dtype == np.uint8:
                column_type = 'SMALLINT'
            elif arr.dtype == np.int32 or arr.dtype == np.uint16:
                column_type = 'INT'
            elif arr.dtype == np.int64 or arr.dtype == np.uint32 or arr.dtype == np.uint64:
                column_type = 'BIGINT'
            elif arr.dtype == np.float32:
                column_type = 'REAL'
            elif arr.dtype == np.float64:
                column_type = 'DOUBLE'
            elif np.issubdtype(arr.dtype, np.str_) or np.issubdtype(arr.dtype, np.unicode_) or np.issubdtype(arr.dtype,
                                                                                                             np.object_):
                column_type = 'STRING'
            else:
                raise Exception('Unsupported dtype: %s' % (str(arr.dtype)))
            column_types.append(column_type)
        query = 'CREATE TABLE %s.%s (' % (
            monet_identifier_escape(schema), monet_identifier_escape(table))
        index = 0
        for key in values.keys():
            query += '%s %s, ' % (monet_identifier_escape(key), column_types[index])
            index += 1
        query = query[:-2] + ");"
        # create the table
        self.execute(query)
        # insert the data into the table
        self.insert(table, values, schema=schema)
        return self

    def _insert_slow(self, table: str, data: Dict[str, np.ndarray], schema: str = 'sys'):
        column_names = data.keys()
        rows = data.values()
        columns = ", ".join([str(i) for i in column_names])
        rows_zipped = list(zip(*rows))
        qmarks = ", ".join(['?'] * len(column_names))
        query = f"insert into {schema}.{table} ({columns}) values ({qmarks})"
        return self.executemany(query, rows_zipped)

    def insert(self, table: str, values: Union[pd.DataFrame, Dict[str, np.ndarray]], schema: str = 'sys'):
        """
        Inserts a set of values into the specified table.

        Args:
            table: The table to insert into
            values: The values. must be either a pandas DataFrame or a dictionary of values.
            schema: The SQL schema to use. If no schema is specified, the "sys" schema is used.
       """
        if isinstance(values, pd.DataFrame):
            prepared = _pandas_to_numpy_dict(values)
        else:
            prepared = values

        for key, value in prepared.items():
            if not isinstance(value, (np.ma.core.MaskedArray, np.ndarray)):  # type: ignore
                prepared[key] = np.array(value)

        if sum(i.dtype.kind not in 'if' for i in prepared.values()):  # type: ignore
            warn(
                "One of the columns you are inserting is not of type int or float which fast append doesn't support. Falling back to regular insert.")
            return self._insert_slow(table, prepared, schema)
        return self.connection.append(schema=schema, table=table, data=prepared)

    def setoutputsize(self, *args, **kwargs) -> None:
        """
        This method would normally set a column buffer size for fetching of large columns.

        MonetDBe-Python does not require this, so calling this function has no effect.
        """
        ...

    def setinputsizes(self, *args, **kwargs) -> None:
        """
        This method would normally be used before the .execute*() method is invoked to reserve memory.

        MonetDBe-Python does not require this, so calling this function has no effect.

        """
        ...

    def commit(self) -> 'Cursor':
        """
        Commit the current pending transaction.

        Returns:
            the current cursor
        """
        if not hasattr(self, 'connection') or not self.connection:
            raise ProgrammingError("no connection to lower level database available")

        self.connection.commit()
        return self

    def transaction(self):
        """
        Start a new transaction

        Returns:
            the current cursor
        """
        return self.execute("BEGIN TRANSACTION")

    def scroll(self, count: int, mode: str = 'relative'):
        """
        Scroll the cursor in the result set to a new position according to mode.

        We dont support scrolling, since the full result is available.
        """
        raise NotImplementedError

    def read_csv(self, table, *args, **kwargs):
        values = pd.read_csv(*args, **kwargs)
        return self.create(table=table, values=values)

    def fetchdf(self) -> pd.DataFrame:
        """
        Fetch all results and return a Pandas DataFrame.

        like .fetchall(), but returns a Pandas DataFrame.
        """
        self._check_connection()
        self._check_result()
        return pd.DataFrame(cast(pd.DataFrame, self.fetchnumpy()))  # cast to make mypy happy

    def fetchmany(self, size=None):
        """
        Fetch the next set of rows of a query result, returning a list of tuples). An empty sequence is returned when
        no more rows are available.

        args:
            size: The number of rows to fetch. Fewer rows may be returned.

        Returns: A number of rows from a query result as a list of tuples

        Raises:
            Error: If the previous call to .execute*() did not produce any result set or no call was issued yet.

        """
        self._check_connection()
        # self._check_result() sqlite test suite doesn't want us to bail out

        if not self.connection.result:
            return []

        if not size:
            size = self.arraysize
        if not self._fetch_generator:
            self._fetch_generator = self.__iter__()

        rows = []
        for i in range(size):
            try:
                rows.append(next(self._fetch_generator))
            except StopIteration:
                break
        return rows

    def fetchone(self) -> Optional[Union['Row', Sequence]]:
        """
        Fetch the next row of a query result set, returning a single tuple, or None when no more data is available.

        Returns:
            One row from a result set.

        Raises:
            Error: If the previous call to .execute*() did not produce any result set or no call was issued yet.

        """
        self._check_connection()
        # self._check_result() sqlite test suite doesn't want us to bail out

        if not self.connection.result:
            return None

        if not self._fetch_generator:
            self._fetch_generator = self.__iter__()
        try:
            return next(self._fetch_generator)  # type: ignore[arg-type]
        except StopIteration:
            return None

    def write_csv(self, table, *args, **kwargs):
        return self.execute(f"select * from {table}").fetchdf().to_csv(*args, **kwargs)

    def __iter_numpy__(self) -> Iterator[Union['Row', Sequence[Any]]]:
        result = self.fetchall()

        for row in result:
            if self.connection.row_factory:  # type: ignore[union-attr]
                yield self.connection.row_factory(cur=self, row=tuple(row))  # type: ignore[union-attr]
            elif self.row_factory:  # Sqlite backwards compatibly
                yield self.row_factory(self, tuple(row))
            else:
                yield tuple(row)

    def fetchall(self) -> List[Union['Row', Sequence]]:
        """
        Fetch all (remaining) rows of a query result, returning them as a list of tuples).

        Returns:
            all (remaining) rows of a query result as a list of tuples

        Raises:
            Error: If the previous call to .execute*() did not produce any result set or no call was issued yet.
        """
        self._check_connection()

        # note (gijs): sqlite test suite doesn't want us to raise exception here, so for now I disable this
        # self._check_result()

        if not self.connection.consistent:
            raise InterfaceError("Tranaction rolled back, state inconsistent")

        if not self.connection.result:
            return []

        rows = [i for i in self]
        self.connection.result = None
        return rows

    def _fetchnumpy_slow(self) -> Mapping[str, np.ndarray]:
        self._check_connection()
        self._check_result()
        all_ = self.fetchall()
        names = (d.name for d in self.description)
        flipped = zip(*all_)
        return {k: v for k, v in zip(names, flipped)}

    def _fetchall_numpy(self):
        result = self.fetchnumpy()
        if result:
            return list(np.vstack(list(result.values())).T)
        return []

    def fetchnumpy(self) -> Mapping[str, np.ndarray]:
        """
        Fetch all results and return a numpy array.

        like .fetchall(), but returns a numpy array.
        """
        from monetdbe._cffi.internal import result_fetch_numpy

        self._check_connection()
        self._check_result()
        return result_fetch_numpy(self.connection.result)  # type: ignore[union-attr]



import unittest

import monetdbe as monetdbe


class TransactionTests(unittest.TestCase):
    def setUp(self):
        self.con = monetdbe.connect()
        self.con.execute("create table test(i int, s text)")

    def tearDown(self):
        self.con.close()

    def test_vanilla(self):
        """This should just work"""
        self.con.execute("insert into test(i, s) values (5, '?%:%(bla)')")

    def test_qmark(self):
        """Question mark style, e.g. ...WHERE name=?"""
        self.con.execute("insert into test(i, s) values (?, ?)", (5, "?%:%(bla)"))

    def test_numeric(self):
        """Numeric, positional style, e.g. ...WHERE name=:1"""
        self.con.execute("insert into test(i, s) values (:1, :2)", (5, "?%:%(bla)"), paramstyle="numeric")

    def test_named(self):
        """Named style, e.g. ...WHERE name=:name"""
        self.con.execute("insert into test(i, s) values (:int, :str)", {'int': 5, 'str': "?%:%(bla)"})

    def test_format(self):
        """ANSI C printf format codes, e.g. ...WHERE name=%s"""
        self.con.execute("insert into test(i, s) values (%s, %s)", (5, "?%:%(bla)"), paramstyle="format")

    def test_pyformat(self):
        """Python extended format codes, e.g. ...WHERE name=%(name)s"""
        self.con.execute("insert into test(i, s) values (%(name)s, %(str)s)", {'name': 5, 'str': "?%:%(bla)"},
                         paramstyle="pyformat")

    def test_wrongparamstyle(self):
        with self.assertRaises(ValueError):
            self.con.execute("insert into test(i, s) values (%(name)s, %(str)s)", {'name': 5, 'str': "?%:%(bla)"},
                             paramstyle="wrong")

    def test_should_fail(self):
        """parameters must be sequence or mapping"""
        with self.assertRaises(ValueError):
            self.con.execute("insert into test(i, s) values (?)", 5)

    @unittest.skip("we don't support multiple open connections yet")
    def test_qmark_multiple_cons(self):
        con1 = monetdbe.connect()
        con1.execute("create table test(i int, s text)")
        con1.execute("insert into test(i, s) values (?, ?)", (5, "bla"))
        con2 = monetdbe.connect()
        con2.execute("create table test(i int, s text)")
        con2.execute("insert into test(i, s) values (?, ?)", (5, "bla"))
        con1.execute("create table test(i int, s text)")
        con1.execute("insert into test(i, s) values (?, ?)", (5, "bla"))



from unittest import TestCase, skip

import numpy
import pandas as pd
import pytest

import monetdbe


class TestmonetdbeBase(TestCase):
    def setUp(self) -> None:
        self.con = monetdbe.connect()

    def test_regular_selection(self):
        con = monetdbe.connect()
        monetdbe.sql('CREATE TABLE pylite00 (i INTEGER)', client=con)
        con.execute('INSERT INTO pylite00 VALUES (1), (2), (3), (4), (5)')
        result = con.execute('SELECT * FROM pylite00').fetchdf()
        assert len(result['i']) == 5, "Incorrect result"

    def test_monetdbe_create(self):
        cur = monetdbe.create('pylite01', {'i': numpy.arange(1000)})
        result = monetdbe.sql('select * from pylite01', client=cur.connection)
        assert len(result['i']) == 1000, "Incorrect result"

    def test_monetdbe_insert(self):
        cur = monetdbe.create('pylite02', {'i': numpy.arange(1000)})
        cur.insert('pylite02', {'i': numpy.arange(1000)})
        result = cur.execute('select * from pylite02').fetchdf()
        assert len(result['i']) == 2000, "Incorrect result"

    def test_monetdbe_create_multiple_columns(self):
        arrays = numpy.arange(100000).reshape((5, 20000))
        cur = monetdbe.create('pylite03',
                              {'i': arrays[0], 'j': arrays[1], 'k': arrays[2], 'l': arrays[3], 'm': arrays[4]})
        result = cur.execute('select * from pylite03').fetchnumpy()
        assert len(result) == 5, "Incorrect amount of columns"
        assert len(result['i']) == 20000, "Incorrect amount of rows"

    def test_sql_types(self):
        con = monetdbe.connect()
        cur = con.execute('CREATE TABLE pylite04_decimal(d DECIMAL(18,3))')
        cur.insert('pylite04_decimal', {'d': numpy.arange(100000).astype(numpy.int64)})
        result = cur.execute('SELECT * FROM pylite04_decimal').fetchdf()
        assert result['d'][0] == 0, "Incorrect result"

        cur.execute('CREATE TABLE pylite04_date(d DATE)')
        cur.execute("INSERT INTO pylite04_date VALUES ('2000-01-01')")
        result = cur.execute('SELECT d FROM pylite04_date').fetchdf()
        assert result['d'][0] == pd.Timestamp('2000-01-01'), "Incorrect result"

    @skip("we don't support multiple open  connections yet")
    def test_connections(self):
        # create two clients
        conn = monetdbe.connect(autocommit=True)
        conn2 = monetdbe.connect()
        # create a table within a transaction in one client
        cur = conn.cursor()
        cur.create('pylite05', {'i': numpy.arange(1000)})

        # check that table was successfully created
        result = monetdbe.sql('SELECT MIN(i) AS minimum FROM pylite05', client=conn)
        assert result['minimum'][0] == 0, "Incorrect result"
        # attempt to query the table from another client
        with pytest.raises(monetdbe.DatabaseError):
            monetdbe.sql('SELECT * FROM pylite05', client=conn2)

        # now commit the table
        monetdbe.sql('COMMIT', client=conn)
        # query the table again from the other client, this time it should be there
        result = monetdbe.sql('SELECT MIN(i) AS minimum FROM pylite05', client=conn2)
        assert result['minimum'][0] == 0, "Incorrect result"

    def test_non_existent_table(self, ):
        # select from non-existent table
        with pytest.raises(monetdbe.DatabaseError):
            monetdbe.sql('select * from nonexistenttable')

    def test_invalid_connection_object(self):
        # invalid connection object
        with pytest.raises(TypeError):
            monetdbe.sql('select * from tables', client=33)

    def test_invalid_colnames(self):
        # invalid colnames
        with pytest.raises(monetdbe.DatabaseError):
            monetdbe.create('pylite06', {33: []})

    def test_empty_colnames(self):
        # empty colnames
        with pytest.raises(monetdbe.DatabaseError):
            monetdbe.create('pylite07', {'': []})

    def test_invalid_key_dict(self):
        # dictionary with invalid keys
        d = dict()
        d[33] = 44
        with pytest.raises(monetdbe.ProgrammingError):
            monetdbe.create('pylite08', d)

    def test_missing_dict_key(self):
        # missing dict key in insert
        cur = monetdbe.create('pylite09', dict(a=[], b=[], c=[]))
        with pytest.raises(monetdbe.DatabaseError):
            cur.insert('pylite09', dict(a=33, b=44))

    def test_bad_column_number(self):
        # too few columns in insert
        cur = monetdbe.create('pylite10', dict(a=[], b=[], c=[]))
        with pytest.raises(monetdbe.DatabaseError):
            cur.insert('pylite10', {'a': [33], 'b': [44]})

    def test_many_sql_statements(self):
        for i in range(5):  # FIXME 1000
            conn = monetdbe.connect()
            cur = conn.execute('CREATE TABLE pylite11 (i INTEGER)')
            cur.insert('pylite11', {'i': numpy.arange(10).astype(numpy.int32)})
            result = cur.execute('SELECT * FROM pylite11').fetchdf()
            assert result['i'][0] == 0, "Invalid result"
            monetdbe.sql('DROP TABLE pylite11', client=conn)
            monetdbe.sql('ROLLBACK', client=conn)
            del conn

    def test_null_string_insertion_bug(self):
        con = monetdbe.connect()
        monetdbe.sql("CREATE TABLE pylite12 (s varchar(2))", client=con)
        monetdbe.insert('pylite12', {'s': ['a', None]}, client=con)
        result = monetdbe.sql("SELECT * FROM pylite12", client=con)
        expected = numpy.ma.masked_array(['a', 'a'], mask=[0, 1])
        numpy.testing.assert_array_equal(result['s'], expected)



from monetdbe import connect
from unittest import TestCase
from pathlib import Path
from tempfile import TemporaryDirectory


class TestCsv(TestCase):
    def test_read_csv(self):
        table = 'test'
        with connect(autocommit=True) as con:
            con.read_csv(
                table=table,
                filepath_or_buffer=Path(__file__).parent / "example.csv",
                names=['i', 's', 'i2', 'f'],
                dtype={"i1": int, 's': str, 'i2': int, 'f': float},
            )
            x = con.execute(f'select * from {table}').fetchall()

    def test_write_csv(self):
        with connect(autocommit=True) as con:
            t = TemporaryDirectory()

            con.write_csv(table='tables', path_or_buf=Path(t.name) / 'output.csv')
